import asyncio
import json
import os
import random
import subprocess
import sys
import traceback
import websockets
from datetime import datetime
from threading import Lock
from typing import Any, Dict, List, Literal, Optional, Union
from uuid import uuid4

import aiohttp
from fastapi import status, APIRouter, BackgroundTasks, Depends, WebSocket, WebSocketDisconnect, HTTPException
from fastapi.responses import JSONResponse
from pydantic import BaseModel, Field

from leaf_playground._type import Singleton
from leaf_playground.core.scene_engine import (
    SceneEngineState,
    SceneObjConfig,
    MetricEvaluatorObjsConfig,
    ReporterObjConfig,
)
from leaf_playground.data.log_body import LogType
from leaf_playground.data.socket_data import SocketData, SocketOperation

from .db import *
from .model import *
from ..hub import Hub
from ..utils import get_local_ip


task_router = APIRouter(prefix="/task")


class LogEvalMetricRecord(BaseModel):
    value: Any = Field(default=...)
    reason: Optional[str] = Field(default=None)
    metric_name: str = Field(default=...)


class LogEvalCompareRecord(LogEvalMetricRecord):
    value: List[str] = Field(default=...)


class TaskCreationPayload(BaseModel):
    project_id: str = Field(default=...)
    scene_obj_config: SceneObjConfig = Field(default=...)
    metric_evaluator_objs_config: MetricEvaluatorObjsConfig = Field(default=...)
    reporter_obj_config: ReporterObjConfig = Field(default=...)

    @property
    def workdir(self):
        return Hub.get_instance().get_project(self.project_id).work_dir


class TaskProxy:
    def __init__(self, tid: str):
        self.id = tid
        self.db: DB = DB.get_instance()

        self._secret_key: str = uuid4().hex
        self._shutdown_event: asyncio.Event = asyncio.Event()
        self._runtime_id: Union[int, str] = None  # pid or docker container name

        asyncio.create_task(self.maybe_shutdown(), name=f"{self.id}_shutdown")

    async def maybe_shutdown(self):
        await self._shutdown_event.wait()
        await asyncio.sleep(3)

        pid = self.runtime_id
        task = await self.db.get_task(self.id)
        if task.runtime_env == "docker":
            subprocess.run(f"docker stop {pid}".split())

    async def get_task(self) -> Task:
        return await self.db.get_task(self.id)

    @property
    def secret_key(self) -> str:
        return self._secret_key

    @property
    def shutdown_event(self) -> asyncio.Event:
        return self._shutdown_event

    @property
    def runtime_id(self) -> int:
        return self._runtime_id

    @runtime_id.setter
    def runtime_id(self, pid: int):
        self._runtime_id = pid


class TaskManager(Singleton):
    def __init__(
        self,
        server_port: int = 8000,
        server_host: str = "127.0.0.1",
        runtime_env: TaskRunTimeEnv = TaskRunTimeEnv.LOCAL,
    ):
        self.hub: Hub = Hub.get_instance()
        self.db: DB = DB.get_instance()
        self.http_session = aiohttp.ClientSession()
        self.result_dir = os.path.join(self.hub.hub_dir, ".leaf_workspace", "results")
        self.tmp_dir = os.path.join(self.hub.hub_dir, ".leaf_workspace", "tmp")
        os.makedirs(self.result_dir, exist_ok=True)
        os.makedirs(self.tmp_dir, exist_ok=True)

        self.server_port = server_port
        self.server_host = server_host
        self._task_ports = set(list(range(10000, 65535))) - {server_port}
        self._port_lock = Lock()

        self.runtime_env = runtime_env

        self._tasks: Dict[str, TaskProxy] = {}

    def acquire_port(self) -> int:
        with self._port_lock:
            port = random.choice(list(self._task_ports))
            self._task_ports -= {port}
        return port

    def release_port(self, port: int) -> None:
        with self._port_lock:
            self._task_ports.add(port)

    async def create_task(self, payload: TaskCreationPayload) -> Task:
        tid = f"task_{payload.project_id}_" + datetime.utcnow().strftime("%Y%m%d%H%M%S") + "_" + uuid4().hex[:8]
        task = Task(
            id=tid,
            project_id=payload.project_id,
            port=self.acquire_port(),
            host=get_local_ip(),
            payload=payload.model_dump_json(by_alias=True),
            runtime_env=self.runtime_env,
            results_dir=os.path.join(self.result_dir, tid),
        )
        self._tasks[tid] = TaskProxy(tid)

        await self.db.insert_task(task)

        if task.runtime_env == "docker":
            self._tasks[tid].runtime_id = self._create_task_in_docker(task)
        else:
            self._tasks[tid].runtime_id = self._create_task_in_local(task)

        return task

    def _create_task_in_local(self, task: Task):
        work_dir = self.hub.get_project(json.loads(task.payload)["project_id"]).work_dir
        process = subprocess.Popen(
            (
                f"{sys.executable} {os.path.join(work_dir, '.leaf', 'app.py')} "
                f"--id {task.id} "
                f"--port {task.port} "
                f"--host {task.host} "
                f"--secret_key {self._tasks[task.id].secret_key} "
                f"--server_url http://{self.server_host}:{self.server_port} "
            ).split()
        )
        return process.pid

    def _create_task_in_docker(self, task: Task):
        container_name = f"leaf-scene-{task.id}"
        image_name = f"leaf-scene-{self.hub.get_project(json.loads(task.payload)['project_id']).name}"
        work_dir = "/app"

        image_check = subprocess.run(f"docker images -q {image_name}", shell=True, capture_output=True, text=True)
        if not image_check.stdout.strip():
            print("Image not found, building Docker image...")
            subprocess.run(f"cd {work_dir} && docker build . -t {image_name}", shell=True)

        subprocess.Popen(
            (
                "docker run --rm "
                f"-p {task.port}:{task.port} "
                f"-v {task.results_dir}:/tmp/result "
                f"--name {container_name} "
                f"{image_name} .leaf/app.py "
                f"--id {task.id} "
                f"--port {task.port} "
                "--host 0.0.0.0 "
                f"--secret_key {self._tasks[task.id].secret_key} "
                f"--server_url http://{self.server_host}:{self.server_port} "
                "--docker"
            ).split()
        )
        return container_name

    async def get_task(self, task_id: str) -> Task:
        return await self.db.get_task(task_id)

    async def get_and_validate_task(self, task_id: str, secret_key: str) -> Task:
        task = await self.get_task(task_id)
        if secret_key != self._tasks[task.id].secret_key:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="incorrect secret key, not allowed to update task status.",
            )
        return task

    async def get_history_tasks(self) -> List[Task]:
        return await self.db.get_task_by_life_cycle(life_cycle=TaskDBLifeCycle.LIVING)

    async def update_task_status(self, task_id: str, task_status: str, secret_key: str):
        await self.get_and_validate_task(task_id, secret_key)
        await self.db.update_task_status(task_id, task_status)
        if task_status in [SceneEngineState.FAILED.value, SceneEngineState.INTERRUPTED.value]:
            self._tasks[task_id].shutdown_event.set()

    async def get_task_agents_connected(self, task_id: str) -> dict:
        task = await self.get_task(task_id)
        url = f"{task.base_http_url}/agents_connected"

        async with self.http_session.get(url=url) as resp:
            if resp.status != 200:
                raise HTTPException(status_code=resp.status, detail=await resp.text())
            return await resp.json()

    async def _control_task_status(self, task_id: str, action: Literal["pause", "resume", "interrupt", "close"]):
        task = await self.get_task(task_id)
        url = f"{task.base_http_url}/{action}"

        async with self.http_session.post(url=url) as resp:
            if resp.status != 200:
                raise HTTPException(status_code=resp.status, detail=await resp.text())

    async def ping_task_service(self, task_id: str) -> str:
        task = await self.get_task(task_id)
        url = f"{task.base_http_url}/hello"

        async with self.http_session.get(url=url) as resp:
            if resp.status != 200:
                raise HTTPException(status_code=resp.status, detail=await resp.text())
            return await resp.text()

    async def pause_task(self, task_id: str):
        await self._control_task_status(task_id, "pause")

    async def resume_task(self, task_id: str):
        await self._control_task_status(task_id, "resume")

    async def interrupt_task(self, task_id: str):
        await self._control_task_status(task_id, "interrupt")

    async def close_task(self, task_id: str):
        await self._control_task_status(task_id, "close")
        self._tasks[task_id].shutdown_event.set()

    async def delete_task(self, task_id: str):
        await self.db.delete_task(task_id)

    async def save_task_results(self, task_id: str):
        task = await self.get_task(task_id)
        url = f"{task.base_http_url}/save"

        async with self.http_session.post(url=url) as resp:
            if resp.status != 200:
                raise HTTPException(status_code=resp.status, detail=await resp.text())

    async def insert_task_log(self, task_id: str, secret_key: str, log: Log):
        await self.get_and_validate_task(task_id, secret_key)
        await self.db.insert_log(log)

    async def update_task_log(self, task_id: str, secret_key: str, log: Log):
        await self.get_and_validate_task(task_id, secret_key)
        await self.db.update_log(log)

    async def websocket_connection(self, task_id: str, websocket: WebSocket, human_id: Optional[str] = None):
        async def _get_and_send_logs():
            nonlocal last_check_time

            logs = await self.db.get_logs_by_tid_with_update_time_constraint(task_id, last_checked_dt=last_check_time)
            for log in logs:
                log_data = log.model_dump(mode="json")
                log_data.update(**log_data.pop("data"))
                if log.log_type == LogType.ACTION:
                    if log_data["references"]:
                        references = []
                        for ref in log_data["references"]:
                            references.append(await self.db.get_message_by_id(ref))
                        log_data["references"] = [ref.data for ref in references]
                    response = await self.db.get_message_by_id(log_data["response"])
                    log_data["response"] = response.data
                socket_operation = (
                    SocketOperation.CREATE if log.created_at == log.last_update else SocketOperation.UPDATE
                )
                if last_check_time is None:
                    socket_operation = SocketOperation.CREATE
                await websocket.send_json(
                    SocketData(data=log_data, operation=socket_operation).model_dump(mode="json")
                )

            if logs:
                last_check_time = logs[-1].db_last_update

        async def _stream_log():
            while True:
                try:
                    await _get_and_send_logs()
                    await asyncio.sleep(0.001)
                except (WebSocketDisconnect, asyncio.CancelledError):
                    break
                except:
                    traceback.print_exc()
                    await websocket.close(code=500, reason=traceback.format_exc())

        last_check_time = None
        await self.get_task(task_id)

        task = await self.get_task(task_id)
        stream_log_task = asyncio.ensure_future(_stream_log())

        if not human_id:
            while True:
                try:
                    text = await websocket.receive_text()
                    if text == "disconnect":
                        raise WebSocketDisconnect()
                except WebSocketDisconnect:
                    stream_log_task.cancel()
                    return
        else:
            async with websockets.connect(f"{task.base_ws_url}/ws/human/{human_id}") as connection:
                while True:
                    try:
                        event = json.loads(await connection.recv())
                        await websocket.send_json(event)
                        if event["event"] == "wait_human_input":
                            human_input = await websocket.receive_text()
                            await connection.send(human_input)
                    except WebSocketDisconnect:
                        stream_log_task.cancel()
                        return
                    except:
                        traceback.print_exc()
                        return

    async def update_task_log_metric_record(
        self, task_id: str, log_id: str, agent_id: str, record: LogEvalMetricRecord
    ):
        task = await self.get_task(task_id)
        url = f"{task.base_http_url}/logs/{log_id}/record/metric/update?agent_id={agent_id}"

        async with self.http_session.post(
            url=url, headers={"Content-Type": "application/json"}, json=record.model_dump(mode="json", by_alias=True)
        ) as resp:
            if resp.status != 200:
                raise HTTPException(status_code=resp.status, detail=await resp.text())

    async def update_task_log_compare_record(self, task_id: str, log_id: str, record: LogEvalCompareRecord):
        task = await self.get_task(task_id)
        url = f"{task.base_http_url}/logs/{log_id}/record/compare/update"

        async with self.http_session.post(
            url=url, headers={"Content-Type": "application/json"}, json=record.model_dump(mode="json", by_alias=True)
        ) as resp:
            if resp.status != 200:
                raise HTTPException(status_code=resp.status, detail=await resp.text())

    async def insert_task_message(self, task_id: str, secret_key: str, message: Message):
        await self.get_and_validate_task(task_id, secret_key)
        await self.db.insert_message(message)

    @classmethod
    async def http_get_task_by_id(cls, task_id: str) -> Task:
        try:
            db: DB = cls.get_instance()
        except KeyError as e:
            raise HTTPException(status_code=status.HTTP_503_SERVICE_UNAVAILABLE, detail="service not ready.")

        return await db.get_task(task_id)


@task_router.post("/create", response_model=Task, response_model_include={"id"})
async def create_task(
    payload: TaskCreationPayload, task_manager: TaskManager = Depends(TaskManager.get_instance)
) -> Task:
    return await task_manager.create_task(payload)


@task_router.get("/{task_id}/hello")
async def ping_task_service(task_id: str, task_manager: TaskManager = Depends(TaskManager.get_instance)):
    return await task_manager.ping_task_service(task_id)


@task_router.post("/{task_id}/pause")
async def pause_task(task_id: str, task_manager: TaskManager = Depends(TaskManager.get_instance)):
    await task_manager.pause_task(task_id)


@task_router.post("/{task_id}/resume")
async def resume_task(task_id: str, task_manager: TaskManager = Depends(TaskManager.get_instance)):
    await task_manager.resume_task(task_id)


@task_router.post("/{task_id}/interrupt")
async def interrupt_task(task_id: str, task_manager: TaskManager = Depends(TaskManager.get_instance)):
    await task_manager.interrupt_task(task_id)


@task_router.post("/{task_id}/close")
async def close_task(task_id: str, task_manager: TaskManager = Depends(TaskManager.get_instance)):
    await task_manager.close_task(task_id)


@task_router.post("/{task_id}/delete")
async def delete_task(
    task_id: str, background_tasks: BackgroundTasks, task_manager: TaskManager = Depends(TaskManager.get_instance)
):
    background_tasks.add_task(task_manager.delete_task, task_id)


@task_router.get("/{task_id}/agents_connected", response_class=JSONResponse)
async def get_task_agents_connected(
    task_id: str, task_manager: TaskManager = Depends(TaskManager.get_instance)
) -> JSONResponse:
    return JSONResponse(content=await task_manager.get_task_agents_connected(task_id))


@task_router.get("/{task_id}/results_dir", response_class=JSONResponse)
async def get_task_results_dir(task: Task = Depends(TaskManager.http_get_task_by_id)) -> JSONResponse:
    return JSONResponse(content={"results_dir": task.results_dir})


@task_router.get("/{task_id}/payload", response_model=TaskCreationPayload)
async def get_task_payload(task: Task = Depends(TaskManager.http_get_task_by_id)) -> TaskCreationPayload:
    return TaskCreationPayload(**json.loads(task.payload))


@task_router.get("/{task_id}/status", response_class=JSONResponse)
async def get_task_status(task: Task = Depends(TaskManager.http_get_task_by_id)) -> JSONResponse:
    return JSONResponse(content={"status": task.status})


@task_router.patch("/{task_id}/status")
async def update_task_status(
    task_id: str, task_status: str, secret_key: str, task_manager: TaskManager = Depends(TaskManager.get_instance)
):
    await task_manager.update_task_status(task_id=task_id, task_status=task_status, secret_key=secret_key)


@task_router.post("/{task_id}/save")
async def save_task_results(task_id: str, task_manager: TaskManager = Depends(TaskManager.get_instance)):
    await task_manager.save_task_results(task_id)


@task_router.get("/history", response_class=JSONResponse)
async def get_history_tasks(task_manager: TaskManager = Depends(TaskManager.get_instance)):
    tasks = await task_manager.get_history_tasks()
    return JSONResponse(
        content=[
            task.model_dump(mode="json", include={"id", "project_id", "status", "created_at", "results_dir"})
            for task in tasks
        ]
    )


@task_router.post("/{task_id}/logs/insert")
async def insert_log(
    task_id: str, secret_key: str, log: Log, task_manager: TaskManager = Depends(TaskManager.get_instance)
):
    await task_manager.insert_task_log(task_id, secret_key, log)


@task_router.patch("/{task_id}/logs/update")
async def update_log(
    task_id: str, secret_key: str, log: Log, task_manager: TaskManager = Depends(TaskManager.get_instance)
):
    await task_manager.update_task_log(task_id, secret_key, log)


@task_router.websocket("/{task_id}/logs/ws")
async def stream_logs(
    task_id: str, websocket: WebSocket, task_manager: TaskManager = Depends(TaskManager.get_instance)
):
    await websocket.accept()
    await task_manager.websocket_connection(task_id, websocket)


@task_router.websocket("/{task_id}/human/{agent_id}/ws")
async def human_connection(
    task_id: str, agent_id: str, websocket: WebSocket, task_manager: TaskManager = Depends(TaskManager.get_instance)
):
    await websocket.accept()
    await task_manager.websocket_connection(task_id, websocket, agent_id)


# TODO: support updating log records after task finished and task service closed.


@task_router.post("/{task_id}/logs/{log_id}/record/metric/update")
async def update_metric_record(
    task_id: str,
    log_id: str,
    agent_id: str,
    record: LogEvalMetricRecord,
    task_manager: TaskManager = Depends(TaskManager.get_instance),
):
    await task_manager.update_task_log_metric_record(task_id, log_id, agent_id, record)


@task_router.post("/{task_id}/logs/{log_id}/record/compare/update")
async def update_compare_record(
    task_id: str,
    log_id: str,
    record: LogEvalCompareRecord,
    task_manager: TaskManager = Depends(TaskManager.get_instance),
):
    await task_manager.update_task_log_compare_record(task_id, log_id, record)


@task_router.post("/{task_id}/messages/insert")
async def insert_message(
    task_id: str, secret_key: str, message: Message, task_manager: TaskManager = Depends(TaskManager.get_instance)
):
    await task_manager.insert_task_message(task_id, secret_key, message)


__all__ = ["task_router", "TaskCreationPayload", "TaskManager", "LogEvalMetricRecord", "LogEvalCompareRecord"]



from typing import Generic, TypeVar

from pydantic import BaseModel
from pydantic._internal._model_construction import ModelMetaclass


T = TypeVar("T")


class Immutable(Generic[T]):
    def __init__(self, value):
        self._value = value

        def __setattr__(cls, __name, __value):
            raise AttributeError("You are changing an immutable variable")

        self.__setattr__ == __setattr__

    def __get__(self, instance, owner):
        return self._value

    def __set__(self, instance, value):
        raise AttributeError("You are changing an immutable variable")

    def __getattr__(self, item):
        try:
            return super().__getattr__(item)
        except:
            return getattr(self._value, item)

    def __getattribute__(self, item):
        try:
            return super().__getattribute__(item)
        except:
            return getattr(self._value, item)

    def __getitem__(self, item):
        return self._value.__getitem__(item)

    def __setitem__(self, key, value):
        raise AttributeError("You are changing an immutable variable")

    def __iter__(self):
        return self._value

    def __next__(self):
        return self._value.__next__()

    def __bool__(self):
        return bool(self._value)

    def __hash__(self):
        return self.__hash__()


class SingletonMetaClass(type):
    _instances: dict = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class Singleton(metaclass=SingletonMetaClass):
    @classmethod
    def get_instance(cls):
        try:
            return cls._instances[cls]
        except KeyError:
            raise KeyError(
                f"class [{cls.__name__}] not instantiate yet, please create one instance before calling this method."
            )


class SingletonBaseModelMetaclass(ModelMetaclass):
    _instances: dict = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class SingletonBaseModel(BaseModel, metaclass=SingletonBaseModelMetaclass):
    @classmethod
    def get_instance(cls):
        try:
            return cls._instances[cls]
        except KeyError:
            raise KeyError(
                f"class [{cls.__name__}] not instantiate yet, please create one instance before calling this method."
            )


__all__ = ["Immutable", "Singleton", "SingletonBaseModel"]



import json
from abc import ABC
from typing import Literal, Optional, Union, Any

from pydantic import Field
from typing_extensions import Annotated

from .base import Data


class Media(Data, ABC):
    display_text: Optional[str] = Field(default=None)
    type: Literal["media"] = Field(default="media")

    def model_post_init(self, __context: Any) -> None:
        if not self.display_text:
            self.display_text = self._generate_display_text()

    def _generate_display_text(self) -> str:
        return None

    def __str__(self):
        return self.display_text


class Text(Media):
    text: str = Field(default=..., frozen=True)
    type: Literal["text"] = Field(default="text")

    def _generate_display_text(self) -> str:
        return self.text[:256]

    def __repr__(self):
        return f"<Text {self.text}>"


class Json(Media):
    data: Union[dict, list] = Field(default={}, frozen=True)
    type: Literal["json"] = Field(default="json")

    def _generate_display_text(self) -> str:
        if isinstance(self.data, dict) and self.data.get("text") is not None:
            return self.data["text"][:256]
        return "Json Object" if isinstance(self.data, dict) else "Json Array"

    def __repr__(self):
        return f"<Json {json.dumps(self.data, ensure_ascii=False, indent=2)}>"


class Image(Media):
    url: str = Field(default=..., frozen=True)
    type: Literal["image"] = Field(default="image")

    def _generate_display_text(self) -> str:
        return f"![Image]({self.url})"

    def __repr__(self):
        return f"<Image {self.url}>"


class Audio(Media):
    type: Literal["audio"] = Field(default="audio")


class Video(Media):
    type: Literal["video"] = Field(default="video")


MediaType = Annotated[Union[Text, Json, Audio, Image, Video], Field(discriminator="type")]


__all__ = ["Media", "Text", "Json", "Image", "Audio", "Video", "MediaType"]



import json
from typing import List, Set

from pydantic import BaseModel


def _get_value(data: dict, field_chain: List[str]):
    field = field_chain.pop(0)
    if field.isdigit():
        field = int(field)
    value = data[field]
    if field_chain:
        return _get_value(value, field_chain)
    return value


class Data(BaseModel):
    """
    Super class for all data structures that can be observed by agents.
    """

    def format(self, template: str, fields: Set[str]) -> str:
        """
        Using a template to formate values of all the selected fields into a beautiful and readable string.

        :param template: a template string used to formate values
        :type template: str
        :param fields: the selected fields whose values are required by the given template
        :type fields: List[str]
        :return: a formatted string
        :rtype: str
        """
        raw_data = json.loads(self.model_dump_json())
        data = dict()
        for field in fields:
            field_chain = field.split(".")
            field_key = "_".join(field_chain)
            data[field_key] = _get_value(raw_data, field_chain)

        return template.format(**data)


__all__ = ["Data"]



from typing import Literal

import altair as alt
from pandas import DataFrame

from .base_vega_chart import BaseVegaChart


class GroupedBar(BaseVegaChart):
    def __init__(self, mode: Literal["percent", "value"] = "value", max_value=1.0):
        self.mode = mode
        self.max_value = max_value

    def generate(self, data: DataFrame) -> dict:
        # use the data format [{agent: gpt4, metric: accuracy, value: 0.81}, ...]

        fmt = ".2f" if self.mode == "value" else ".1%"

        chart = (
            alt.Chart(data)
            .mark_bar(cornerRadius=5, height={"band": 0.6})
            .encode(
                y=alt.X("metric:N").axis(labelAngle=0, title=None),
                x=alt.Y("value:Q").axis(title="value", format=fmt).scale(domain=(0, self.max_value)),
                color=alt.Color("metric:N"),
            )
        )

        text = chart.mark_text(align="left", dx=3).encode(text=alt.Text("value:Q", format=fmt))

        return (chart + text).facet(row=alt.Row("agent:N", header=alt.Header(title=None))).to_dict()



from leaf_playground.core.workers import Logger
from leaf_playground.core.scene import Scene
from leaf_playground.core.scene_definition import SceneConfig
from leaf_playground.data.log_body import ActionLogBody

from .scene_definition import *


class {{ cookiecutter.project_name_camel_case }}SceneLogBody(ActionLogBody):
    pass


{{ cookiecutter.project_name_camel_case }}SceneConfig = SceneConfig.create_config_model(
    SCENE_DEFINITION,
)


class {{ cookiecutter.project_name_camel_case }}Scene(
    Scene,
    scene_definition=SCENE_DEFINITION,
    log_body_class={{ cookiecutter.project_name_camel_case }}SceneLogBody
):
    config_cls = {{ cookiecutter.project_name_camel_case }}SceneConfig
    config: config_cls

    def __init__(self, config: config_cls, logger: Logger):
        super().__init__(config=config, logger=logger)
        # TODO: additional initialization logic here

    async def _run(self):
        # TODO: scene flow here
        pass


__all__ = [
    "{{ cookiecutter.project_name_camel_case }}SceneConfig",
    "{{ cookiecutter.project_name_camel_case }}Scene"
]



import socket


def get_local_ip() -> str:
    try:
        # Create a temporary socket
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
            # Connect to a public DNS server (Google's)
            s.connect(("8.8.8.8", 80))
            # Get the local IP address from the socket
            local_ip = s.getsockname()[0]
            return local_ip
    except Exception:
        return "127.0.0.1"


__all__ = ["get_local_ip"]



from typing import List, Optional

from leaf_playground.core.scene_agent import SceneStaticAgentConfig, SceneStaticAgent
from leaf_playground.data.profile import Profile
from leaf_playground.data.media import Text

from ..scene_definition import ExaminerSample, SCENE_DEFINITION
from ..dataset_util import *

ROLE_DEFINITION = SCENE_DEFINITION.get_role_definition("examiner")

ExaminerConfig = SceneStaticAgentConfig.create_config_model(ROLE_DEFINITION)


class Examiner(SceneStaticAgent, role_definition=ROLE_DEFINITION, cls_description=ROLE_DEFINITION.description):
    config_cls = ExaminerConfig
    config: config_cls

    def __init__(self, config: config_cls):
        super().__init__(config=config)

        self._cur = 0
        self._questions = []
        self._ds_config: DatasetConfig = None

    async def prepare_samples(self, ds_config: DatasetConfig) -> None:
        self._cur = 0
        self._questions = prepare_samples(ds_config)
        self._ds_config = ds_config

    async def send_sample(self, receivers: List[Profile]) -> ExaminerSample:
        sample = ExaminerSample(
            sender=self.profile,
            receivers=receivers,
            content=Text(
                text=self._questions[self._cur][QUESTION_COL], display_text=self._questions[self._cur][QUESTION_COL]
            ),
            sample_id=self._cur,
        )
        self._cur += 1
        return sample

    def check_examine_finish(self) -> bool:
        return self._cur >= len(self._questions)

    def get_golden_answer(self, sample_id: int) -> Optional[str]:
        return self._questions[sample_id][ANSWER_COL]


__all__ = ["ExaminerConfig", "Examiner"]



from sys import _getframe
from typing import Any, Type

from pydantic import create_model, field_serializer, BaseModel, Field, PrivateAttr

from ...._config import _Config
from ....data.environment import EnvironmentVariable


class EnvVarDefinition(BaseModel):
    name: str = Field(default=...)
    description: str = Field(default=...)
    env_var_cls: Type[EnvironmentVariable] = Field(default=EnvironmentVariable)

    def model_post_init(self, __context: Any) -> None:
        if not issubclass(self.env_var_cls, EnvironmentVariable):
            raise ValueError(f"env_var_cls [{self.env_var_cls.__name__}] should be a subclass of EnvironmentVariable")

    @field_serializer("env_var_cls")
    def serialize_type(self, env_var_cls: Type[EnvironmentVariable], _info) -> str:
        return env_var_cls.__name__


class EnvVarConfig(_Config):
    _env_var_definition: EnvVarDefinition = PrivateAttr(default=None)

    def __init_subclass__(cls, _env_var_definition: EnvVarDefinition, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._env_var_definition = _env_var_definition

    @property
    def env_var_definition(self) -> EnvVarDefinition:
        return self._env_var_definition

    @classmethod
    def create_config_model(cls, env_var_definition: EnvVarDefinition) -> Type["EnvironmentVariable"]:
        model_name = "".join([each.capitalize() for each in env_var_definition.name.split("_")]) + cls.__name__
        module = _getframe(1).f_globals["__name__"]
        cls_kwargs = {"_env_var_definition": env_var_definition}
        fields = {
            f_name: (f_info.annotation, f_info)
            for f_name, f_info in env_var_definition.env_var_cls.model_fields.items()
            if f_name not in ["name", "description"] and not f_info.exclude
        }
        return create_model(
            __model_name=model_name, __module__=module, __base__=cls, __cls_kwargs__=cls_kwargs, **fields
        )

    def initiate_env_var(self) -> EnvironmentVariable:
        return self.env_var_definition.env_var_cls(
            name=self.env_var_definition.name,
            description=self.env_var_definition.description,
            **self.model_dump(mode="json", by_alias=True),
        )


__all__ = ["EnvVarDefinition", "EnvVarConfig"]



from enum import Enum
from inspect import signature, Signature, Parameter
from functools import partial
from hashlib import md5
from sys import _getframe
from typing import Any, Dict, List, Literal, Optional, Tuple, Type, Union

from pydantic import create_model, field_serializer, BaseModel, Field, PrivateAttr

from ...._config import _Config
from ....data.base import Data
from ....data.message import Message
from ....utils.import_util import DynamicFn

_METRIC_MODELS = {}


class ValueDType(Enum):
    INT = int
    FLOAT = float
    BOOLEAN = bool


VALUE_DETYPE_2_DEFAULT_VALUE = {ValueDType.INT: 0, ValueDType.FLOAT: 0.0, ValueDType.BOOLEAN: False}


class DisplayType(Enum):
    FIVE_STARTS_RATE = "FiveStarsRate"
    NUMBER_INPUT = "NumberInput"
    BOOLEAN_RADIO = "BooleanRadio"


class _RecordData(Data):
    value: Any = Field(default=...)
    evaluator: str = Field(default=...)
    display_type: DisplayType = Field(default=...)
    reason: Optional[str] = Field(default=None)
    misc: Optional[dict] = Field(default=None)
    is_comparison: bool = Field(default=False)
    target_agent: Optional[str] = Field(default=None)


class _MetricData(Data):
    value: Any = Field(default=...)
    records: List[_RecordData] = Field(default=...)
    is_comparison: bool = Field(default=False)
    target_agent: Optional[str] = Field(default=None)


CombinedMetricsData = Dict[
    Literal["metrics", "human_metrics", "merged_metrics"], Dict[str, Union[_MetricData, List[_MetricData]]]
]


class AggregationMethodOutput(BaseModel):
    value: Any = Field(default=...)


class DynamicAggregationFn(DynamicFn):
    @classmethod
    def create_dynamic_fn(cls, fn: Type, default_kwargs: Optional[dict] = None) -> "DynamicAggregationFn":
        new_fn = fn
        new_fn_signature = signature(new_fn)
        if default_kwargs:
            new_fn = partial(fn, **default_kwargs)
            new_fn_signature = signature(new_fn)
            new_fn_signature = new_fn_signature.replace(
                parameters=[p for name, p in new_fn_signature.parameters.items() if name not in default_kwargs]
            )

        required_signature = Signature(
            parameters=[Parameter(name="records", kind=Parameter.POSITIONAL_OR_KEYWORD, annotation=List[_RecordData])],
            return_annotation=AggregationMethodOutput,
        )

        if required_signature != new_fn_signature:
            raise TypeError(
                f"expected signature is {required_signature}, "
                f"got {new_fn_signature} which originally is {signature(fn)}"
            )

        dynamic_fn = super().create_dynamic_fn(fn, default_kwargs)
        return cls(**dynamic_fn.model_dump(mode="json", by_alias=True))


DefaultAggregateMethod = Literal["mean", "min", "max", "median", "sum"]
DynamicAggregationMethod = Union[DefaultAggregateMethod, DynamicAggregationFn]

DEFAULT_AGG_METHODS = {
    "mean": lambda x: AggregationMethodOutput(value=sum([each.value for each in x]) / len(x)),
    "min": lambda x: AggregationMethodOutput(value=min([each.value for each in x])),
    "max": lambda x: AggregationMethodOutput(value=max([each.value for each in x])),
    "median": lambda x: AggregationMethodOutput(value=sorted([each.value for each in x])[len(x) // 2]),
    "sum": lambda x: AggregationMethodOutput(value=sum([each.value for each in x])),
}


class CompareMetricDefinition(BaseModel):
    name: str = Field(default=...)
    description: str = Field(default=...)
    expect_resp_msg_type: Type = Field(default=...)
    is_comparison: Literal[True] = Field(default=True)

    _belonged_action: Optional["leaf_playground.core.scene_info.definitions.action.ActionDefinition"] = PrivateAttr(
        default=None
    )

    @property
    def belonged_action(self) -> "leaf_playground.core.scene_info.definitions.action.ActionDefinition":
        return self._belonged_action

    @property
    def belonged_chain(self):
        return self.belonged_action.belonged_chain + "." + self.name

    @field_serializer("expect_resp_msg_type")
    def serialize_valid_msg_types(self, expect_resp_msg_type: Type, _info):
        return expect_resp_msg_type.__name__

    def model_post_init(self, __context: Any) -> None:
        if not self.expect_resp_msg_type:
            raise ValueError(f"valid_msg_types should not be empty")
        if not issubclass(self.expect_resp_msg_type, Message):
            raise TypeError(f"expect_resp_msg_type must be a subclass of Message")

    def create_data_models(self) -> Tuple[Type[_MetricData], Type[_RecordData]]:
        hash_id = md5(self.model_dump_json().encode()).hexdigest()
        if hash_id in _METRIC_MODELS:
            return _METRIC_MODELS[hash_id]

        base_name = "".join([s.capitalize() for s in self.name.split("_")])
        record_model_name = base_name + "ComparisonRecord"
        metric_model_name = base_name + "Comparison"
        module = _getframe(1).f_globals["__name__"]

        record_model_fields = {
            "value": (List[str], Field(default=[])),
            "display_type": (Literal[None], Field(default=None)),
            "is_comparison": (Literal[self.is_comparison], Field(default=self.is_comparison)),
            "target_agent": (Literal[None], Field(default=None)),
        }

        record_model = create_model(
            __model_name=record_model_name, __module__=module, __base__=_RecordData, **record_model_fields
        )

        metric_model_fields = {
            "records": (List[record_model], Field(default=...)),
            "is_comparison": (Literal[self.is_comparison], Field(default=self.is_comparison)),
            "target_agent": (Literal[None], Field(default=None)),
        }

        metric_model = create_model(
            __model_name=metric_model_name, __module__=module, __base__=_MetricData, **metric_model_fields
        )

        _METRIC_MODELS[hash_id] = (metric_model, record_model)

        return metric_model, record_model


class MetricDefinition(BaseModel):
    name: str = Field(default=...)
    description: str = Field(default=...)
    record_value_dtype: ValueDType = Field(default=...)
    record_display_type: DisplayType = Field(default=...)
    expect_resp_msg_type: Type = Field(default=...)
    agg_method: DynamicAggregationMethod = Field(default=...)
    is_comparison: Literal[False] = Field(default=False)

    _belonged_action: Optional["leaf_playground.core.scene_info.definitions.action.ActionDefinition"] = PrivateAttr(
        default=None
    )

    @property
    def belonged_action(self) -> "leaf_playground.core.scene_info.definitions.action.ActionDefinition":
        return self._belonged_action

    @property
    def belonged_chain(self) -> str:
        return self.belonged_action.belonged_chain + "." + self.name

    def model_post_init(self, __context: Any) -> None:
        if not self.expect_resp_msg_type:
            raise ValueError(f"valid_msg_types should not be empty")
        if not issubclass(self.expect_resp_msg_type, Message):
            raise TypeError(f"expect_resp_msg_type must be a subclass of Message")

    @field_serializer("expect_resp_msg_type")
    def serialize_valid_msg_types(self, expect_resp_msg_type: Type, _info):
        return expect_resp_msg_type.__name__

    @field_serializer("record_value_dtype")
    def serialize_record_value_dtype(self, record_value_dtype: ValueDType, _info):
        return record_value_dtype.name

    def get_value_annotation(self):
        return self.record_value_dtype.value

    def create_data_models(self) -> Tuple[Type[_MetricData], Type[_RecordData]]:
        hash_id = md5(self.model_dump_json().encode()).hexdigest()
        if hash_id in _METRIC_MODELS:
            return _METRIC_MODELS[hash_id]

        base_name = "".join([s.capitalize() for s in self.name.split("_")])
        record_model_name = base_name + "MetricRecord"
        metric_model_name = base_name + "Metric"
        module = _getframe(1).f_globals["__name__"]

        value_annotation = self.get_value_annotation()

        record_model_fields = {
            "value": (value_annotation, Field(default=VALUE_DETYPE_2_DEFAULT_VALUE[self.record_value_dtype])),
            "display_type": (Literal[self.record_display_type], Field(default=self.record_display_type)),
            "is_comparison": (Literal[self.is_comparison], Field(default=self.is_comparison)),
            "target_agent": (str, Field(default=...)),
        }

        record_model = create_model(
            __model_name=record_model_name, __module__=module, __base__=_RecordData, **record_model_fields
        )

        metric_model_fields = {
            "records": (List[record_model], Field(default=...)),
            "is_comparison": (Literal[self.is_comparison], Field(default=self.is_comparison)),
            "target_agent": (str, Field(default=...)),
        }

        metric_model = create_model(
            __model_name=metric_model_name, __module__=module, __base__=_MetricData, **metric_model_fields
        )

        _METRIC_MODELS[hash_id] = (metric_model, record_model)

        return metric_model, record_model


class MetricConfig(_Config):
    enable: bool = Field(default=True)


__all__ = [
    "ValueDType",
    "DisplayType",
    "CombinedMetricsData",
    "AggregationMethodOutput",
    "DynamicAggregationMethod",
    "DynamicAggregationFn",
    "DEFAULT_AGG_METHODS",
    "VALUE_DETYPE_2_DEFAULT_VALUE",
    "CompareMetricDefinition",
    "MetricDefinition",
    "MetricConfig",
]



from collections import defaultdict
from typing import Any, Dict, List, Type

from .chart import Chart
from ..scene_definition import CombinedMetricsData, MetricDefinition, SceneConfig, SceneDefinition
from ..scene_definition.definitions.metric import (
    _RecordData,
    _MetricData,
    DynamicAggregationMethod,
    DEFAULT_AGG_METHODS,
)
from ...data.log_body import LogBody
from ...utils.import_util import dynamically_import_fn


def _agg(records: List[_RecordData], agg_method: DynamicAggregationMethod) -> Any:
    if isinstance(agg_method, str):
        agg_method_ = DEFAULT_AGG_METHODS[agg_method]
    else:
        agg_method_ = dynamically_import_fn(agg_method)
    return agg_method_(records).value


def _win_ratio(ranks: List[List[str]], top_n: int):
    counter = {agent: 0 for agent in ranks[0]}
    for rank in ranks:
        for agent in rank[:top_n]:
            counter[agent] += 1
    return {k: v / len(ranks) for k, v in counter.items()}


class MetricReporter:
    def __init__(self, scene_definition: SceneDefinition, chart_classes: List[Type[Chart]]):
        metric_definitions = {}
        for role in scene_definition.roles:
            for action in role.actions:
                for metric_def in action.metrics or []:
                    metric_definitions[metric_def.belonged_chain] = metric_def
        self.metric_definitions: Dict[str, MetricDefinition] = metric_definitions
        self.records: Dict[str, Dict[str, List[_RecordData]]] = defaultdict(dict)
        self.charts = [chart_cls() for chart_cls in chart_classes]
        self.human_records: Dict[str, Dict[str, _RecordData]] = defaultdict(dict)

    def put_record(self, record: _RecordData, metric_belonged_chain: str, log_id: str):
        if log_id not in self.records[metric_belonged_chain]:
            self.records[metric_belonged_chain][log_id] = []
        self.records[metric_belonged_chain][log_id].append(record)

    def put_human_record(self, record: _RecordData, metric_belonged_chain: str, log_id: str):
        self.human_records[metric_belonged_chain][log_id] = record

    def _cal_record_metric(self, records: List[_RecordData], metric_def: MetricDefinition) -> List[_MetricData]:
        agg_method = metric_def.agg_method
        metric_data_model, _ = metric_def.create_data_models()

        agent2records = defaultdict(list)
        for record in records:
            agent2records[record.target_agent].append(record)

        metrics = []
        for agent_id, agent_records in agent2records.items():
            metrics.append(
                metric_data_model(value=_agg(agent_records, agg_method), records=agent_records, target_agent=agent_id)
            )

        return metrics

    def _cal_compare_metric(self, records: List[_RecordData], metric_def: MetricDefinition) -> _MetricData:
        metric_data_model, _ = metric_def.create_data_models()
        record_values: List[List[str]] = [record.value for record in records]

        num_agents = len(record_values[0])

        result = {"win_ratio": _win_ratio(record_values, top_n=1)}
        if num_agents >= 10:
            result["top3_ratio"] = _win_ratio(record_values, top_n=3)
        if num_agents >= 30:
            result["top10_ratio"] = _win_ratio(record_values, top_n=10)

        return metric_data_model(value=result, records=records)

    def _cal_metrics(self) -> CombinedMetricsData:
        metrics = {}
        human_metrics = {}
        merged_metrics = {}

        for metric_belonged_chain, records in self.records.items():
            metric_def = self.metric_definitions[metric_belonged_chain]
            is_compare = metric_def.is_comparison

            flatten_records = [record for records_ in records.values() for record in records_]

            if not is_compare:
                metrics[metric_belonged_chain] = self._cal_record_metric(flatten_records, metric_def)
            else:
                metrics[metric_belonged_chain] = self._cal_compare_metric(flatten_records, metric_def)

        for metric_belonged_chain, records in self.human_records.items():
            metric_def = self.metric_definitions[metric_belonged_chain]
            is_compare = metric_def.is_comparison

            if not is_compare:
                human_metrics[metric_belonged_chain] = self._cal_record_metric(list(records.values()), metric_def)
            else:
                human_metrics[metric_belonged_chain] = self._cal_compare_metric(list(records.values()), metric_def)

        for metric_belonged_chain, records in self.records.items():
            metric_def = self.metric_definitions[metric_belonged_chain]
            human_records = self.human_records[metric_belonged_chain]
            is_compare = metric_def.is_comparison

            merged_records = []
            for log_id, record in records.items():
                if log_id in human_records:
                    merged_records.append(human_records[log_id])
                else:
                    merged_records.extend(record)

            if not is_compare:
                merged_metrics[metric_belonged_chain] = self._cal_record_metric(merged_records, metric_def)
            else:
                merged_metrics[metric_belonged_chain] = self._cal_compare_metric(merged_records, metric_def)

        return {"metrics": metrics, "human_metrics": human_metrics, "merged_metrics": merged_metrics}

    @property
    def metrics_data(self) -> CombinedMetricsData:
        return self._cal_metrics()

    def generate_reports(
        self,
        scene_config: SceneConfig,
        evaluator_configs: List["leaf_playground.core.workers.MetricEvaluatorConfig"],
        logs: List[LogBody],
    ):
        metrics = self.metrics_data

        charts = {
            chart.chart_name: chart.generate(metrics, scene_config, evaluator_configs, logs) for chart in self.charts
        }
        charts = {chart_name: chart for chart_name, chart in charts.items() if chart is not None}

        return metrics, charts


__all__ = ["MetricReporter"]



from typing import Optional

from pydantic import Field

from .base import Data


class Tool(Data):
    pass


class HFTool(Tool):
    repo_id: str = Field(default=...)
    model_repo_id: Optional[str] = Field(default=None)
    token: Optional[str] = Field(default=None)
    remote: bool = Field(default=False)


__all__ = ["Tool", "HFTool"]



from typing import Dict, List, Optional, Union, Any

from pydantic import Field

from .base import AIBackend, AIBackendConfig
from .._config import _Config
from ..data.tool import HFTool


class HFLocalModelConfig(_Config):
    model_name_or_path: str = Field(default=...)
    tokenizer_name_or_path: Optional[str] = Field(default=None)
    revision: str = Field(default="main")
    low_cpu_mem_usage: bool = Field(default=True)
    torch_dtype: str = Field(default="float16", pattern="(float16|bfloat16)")
    device: Optional[Union[int, str]] = Field(default=None)
    max_memory: Optional[Dict[Union[str, int], str]] = Field(default=None)
    device_map: Union[str, Dict[str, Union[int, str]]] = Field(default="auto")
    use_fast_tokenizer: bool = Field(default=False)
    trust_remote_code: bool = Field(default=False)
    use_safetensors: bool = Field(default=False)

    def prepare_model_tokenizer(self):
        import torch
        from transformers import AutoTokenizer, AutoModelForCausalLM

        tokenizer = AutoTokenizer.from_pretrained(
            self.config.tokenizer_name_or_path or self.config.model_name_or_path,
            use_fast=self.config.use_fast_tokenizer,
            padding_side="left",
            truncation_side="left",
            trust_remote_code=self.config.trust_remote_code,
        )
        # TODO: specify pad_token_id to eos_token_id if it not exists using new way that transformers now supports

        max_memory = self.config.max_memory
        if max_memory:
            max_memory = {(eval(k) if isinstance(k, str) else k): v for k, v in max_memory.items()}

        model = AutoModelForCausalLM.from_pretrained(
            pretrained_model_name_or_path=self.config.model_name_or_path,
            torch_dtype=getattr(torch, self.config.torch_dtype),
            device_map=self.config.device_map,
            max_memory=max_memory,
            low_cpu_mem_usage=self.config.low_cpu_mem_usage,
            revision=self.config.revision,
            trust_remote_code=self.config.trust_remote_code,
        )

        return model, tokenizer


class HFRemoteEndpointConfig(_Config):
    url_endpoint: str = Field(default=...)
    token: Optional[str] = Field(default=None)


class OpenAIConfig(_Config):
    api_key: Optional[str] = Field(default=None)
    model: str = Field(default="text-davinci-003")


class HFAgentBackendConfig(AIBackendConfig):
    remote_endpoint_config: Optional[HFRemoteEndpointConfig] = Field(default=None)
    local_model_config: Optional[HFLocalModelConfig] = Field(default=None)
    openai_config: Optional[OpenAIConfig] = Field(default=None)
    chat_prompt_template: Optional[str] = Field(default=None)
    run_prompt_template: Optional[str] = Field(default=None)
    additional_tools: Optional[List[HFTool]] = Field(default=None)

    def model_post_init(self, __context: Any) -> None:
        if not self.remote_endpoint_config and not self.local_model_config and not self.openai_config:
            raise ValueError(
                "Must provide one of the `remote_endpoint_config`, `local_model_config`, or `openai_config` arguments"
            )
        if sum([bool(self.remote_endpoint_config), bool(self.local_model_config), bool(self.openai_config)]) > 1:
            raise ValueError(
                "`remote_endpoint_config`, `local_model_config` and `openai_config` are mutually exclusive"
            )

    def create_hf_agent(self):
        from transformers import HfAgent, Tool

        payload = {"chat_prompt_template": self.chat_prompt_template, "run_prompt_template": self.run_prompt_template}
        if self.additional_tools:
            payload["additional_tools"] = [Tool.from_hub(**tool.model_dump()) for tool in self.additional_tools]
        if self.remote_endpoint_config:
            payload.update(**self.hf_endpoint.model_dump())
        elif self.local_model_config:
            model, tokenizer = self.local_model_config.prepare_model_tokenizer()
            payload.update(model=model, tokenizer=tokenizer)
        else:
            payload.update(**self.openai_config.model_dump())

        return HfAgent(**payload)


class HFPipelineBackendConfig(HFLocalModelConfig):
    def prepare_pipeline(self):
        from transformers import TextGenerationPipeline

        model, tokenizer = self.prepare_model_tokenizer()
        return TextGenerationPipeline(model=model, tokenizer=tokenizer)


class HFAgentBackend(AIBackend):
    config_cls = HFAgentBackendConfig
    config: config_cls

    def __init__(self, config: config_cls):
        super().__init__(config=config)
        self.hf_agent = self.config.create_hf_agent()


class HFPipelineBackend(AIBackend):
    config_cls = HFPipelineBackendConfig
    config: config_cls

    def __init__(self, config: config_cls):
        super().__init__(config=config)
        self.pipeline = self.config.prepare_pipeline()


__all__ = [
    "HFLocalModelConfig",
    "HFRemoteEndpointConfig",
    "OpenAIConfig",
    "HFAgentBackendConfig",
    "HFAgentBackend",
    "HFPipelineBackendConfig",
    "HFPipelineBackend",
]



import json
import os
import re
import time
from typing import Any, Optional, Union

import jinja2
from pydantic import BaseModel, Field, PrivateAttr
from sglang import function, system, user, assistant, gen, set_default_backend, OpenAI
from sglang.lang.interpreter import ProgramState, SglGen


prompt_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "prompts")


def check_is_leaf_project(directory: str):
    if not os.path.exists(os.path.join(directory, ".leaf")):
        raise ValueError(f"{directory} must be a project directory that generated by leaf-playground CLI")


def add_and_show_message(
    state: ProgramState, msg_wrapper: Union[system, user, assistant], msg: Union[str, SglGen], show: bool = True
):
    if isinstance(msg, SglGen) and not msg.name:
        raise ValueError("Must set name to SglGen instance")
    state += msg_wrapper(msg)
    if msg_wrapper != system and show:
        if isinstance(msg, str):
            message = state.messages()[-1]
            print(f"{message['role'].title()}: {message['content']}", end="\n\n")
        else:
            print(f"{'User' if msg_wrapper == user else 'Assistant'}: ", end="")
            for text in state.text_iter(var_name=msg.name):
                print(text, end="", flush=True)
            print()


def load_prompt(*rel_path: str):
    with open(os.path.join(prompt_dir, *rel_path), "r", encoding="utf-8") as f:
        return f.read()


def load_code(reference_file: Optional[str], default: Optional[str]):
    if not reference_file:
        return default
    with open(reference_file, "r", encoding="utf-8") as f:
        return f.read()


def extract_code_and_save(code: str, file_name: str, save_dir: str):
    pattern = r"```python\n# " + file_name + r".py(.*?)```"
    match = re.search(pattern, code, re.DOTALL)
    if match:
        code = match.group(1).strip()

    with open(os.path.join(save_dir, f"{file_name}.py"), "w", encoding="utf-8") as f:
        f.write(code)


class PipelineConfig(BaseModel):
    target_project: str = Field(default=...)
    reference_project: Optional[str] = Field(default=None)
    openai_api_key: Optional[str] = Field(default=None)

    definition_completion: bool = Field(default=True)
    agent_completion: bool = Field(default=True)
    scene_completion: bool = Field(default=True)

    _target_project_config: dict = PrivateAttr(default=None)
    _target_pkg: str = PrivateAttr(default=None)
    _reference_pkg: str = PrivateAttr(default=None)

    def model_post_init(self, __context: Any) -> None:
        if self.openai_api_key is None:
            self.openai_api_key = os.environ.get("OPENAI_API_KEY", None)
        if not self.openai_api_key:
            raise ValueError(
                "Must provide the `openai_api_key` argument or the `OPENAI_API_KEY` environment variable。"
            )

        check_is_leaf_project(self.target_project)
        if self.reference_project:
            check_is_leaf_project(self.reference_project)

        self._target_project_config = self.load_project_config(self.target_project)
        self._target_pkg = self.get_pkg_path(self.target_project)
        if self.reference_project:
            self._reference_pkg = self.get_pkg_path(self.reference_project)

    @staticmethod
    def load_project_config(project_path: str):
        with open(os.path.join(project_path, ".leaf", "project_config.json"), "r", encoding="utf-8") as f:
            proj_config = json.load(f)
        return proj_config

    @staticmethod
    def get_pkg_path(project_path: str):
        proj_config = PipelineConfig.load_project_config(project_path)
        return os.path.join(project_path, proj_config["name"])

    @property
    def target_project_config(self):
        return self._target_project_config

    @property
    def target_pkg_path(self):
        return self._target_pkg

    @property
    def reference_pkg_path(self):
        return self._reference_pkg


class Pipeline:
    def __init__(self, config: PipelineConfig):
        set_default_backend(OpenAI("gpt-4-1106-preview", api_key=config.openai_api_key))

        self.config = config

    def _definition_completion(self):
        initial_msg = jinja2.Template(load_prompt("system", "product_manager.prompt")).render(
            leaf_playground_overview=load_prompt("docs", "overview.prompt"),
            definition_doc=jinja2.Template(load_prompt("docs", "definition.prompt")).render(
                data_message_module_reference=load_prompt("docs", "references", "data_message_module.prompt"),
                data_environment_module_reference=load_prompt("docs", "references", "data_environment_module.prompt"),
                core_definition_module_reference=load_prompt("docs", "references", "core_definition_module.prompt"),
            ),
            scene_definition_example=load_code(
                reference_file=(
                    None
                    if not self.config.reference_pkg_path
                    else os.path.join(self.config.reference_pkg_path, "scene_definition.py")
                ),
                default=None if self.config.reference_pkg_path else load_prompt("examples", "scene_definition.py"),
            ),
            project_name=self.config.target_project_config["name"],
            project_description=input(
                jinja2.Template(load_prompt("user_inputs", "ask_for_project_description.prompt")).render(
                    project_name=self.config.target_project_config["name"]
                )
            ),
        )

        _finished = False

        @function
        def _complete(s, initial_msg):
            nonlocal _finished

            add_and_show_message(s, user, initial_msg, False)
            system_msg = "阶段：第一阶段；要求：{0}"
            while True:
                add_and_show_message(
                    s,
                    system,
                    system_msg.format(
                        "先响应用户的回复，然后一步步地思考（think step by step）并使用自然语言将需求细化，暂时不要提问"
                    ),
                )
                add_and_show_message(s, assistant, gen("detailed_solution", max_tokens=4096))
                add_and_show_message(
                    s,
                    system,
                    system_msg.format("判断是否需要向用户提问或寻求改进意见，直接回答“是”或“否”，不要说其他的话"),
                )
                add_and_show_message(s, assistant, gen("judgement", max_tokens=2, temperature=0.0), False)
                if "否" in s.get_var("judgement"):
                    break
                add_and_show_message(
                    s,
                    system,
                    system_msg.format(
                        "先告知用户你接下来要做的事情，然后分点列出你需要向用户提问的全部问题或向用户寻求改进意见"
                    ),
                )
                add_and_show_message(s, assistant, gen("questions", max_tokens=128))
                add_and_show_message(s, user, input())
                add_and_show_message(
                    s,
                    system,
                    system_msg.format("根据用户回复判断是否需要修改需求细节，直接回答“是”或“否”，不要说其他的话"),
                )
                add_and_show_message(s, assistant, gen("judgement", max_tokens=2, temperature=0.0), False)
                if "否" in s.get_var("judgement"):
                    break
            system_msg = "阶段：第二阶段；要求：{0}"
            while True:
                add_and_show_message(
                    s,
                    system,
                    system_msg.format(
                        "结合用户在一开始给你的指南并根据最新细化的需求，生成 scene_definition.py，然后询问用户是否有需要补充或修改的"
                    ),
                )
                add_and_show_message(s, assistant, gen("scene_definition.py", max_tokens=4096))
                add_and_show_message(s, user, input())
                add_and_show_message(
                    s,
                    system,
                    "根据用户的回复判断你是否需要对 scene_definition.py 进行修改？直接回答“是”或“否”，不要说其他的话",
                )
                add_and_show_message(s, assistant, gen("judgement", max_tokens=2, temperature=0.0), False)
                if "否" in s.get_var("judgement"):
                    break

            extract_code_and_save(s.get_var("scene_definition.py"), "scene_definition", self.config.target_pkg_path)

            _finished = True

        _complete.run(initial_msg, temperature=0.7, stream=True)

        while not _finished:
            time.sleep(0.01)

    def _agent_completion(self):
        target_scene_definition = load_code(os.path.join(self.config.target_pkg_path, "scene_definition.py"), None)
        if not target_scene_definition:
            raise NotImplementedError("must implement scene_definition.py before complete agents' code")

        initial_msg = jinja2.Template(load_prompt("system", "base_agent_developer.prompt")).render(
            leaf_playground_overview=load_prompt("docs", "overview.prompt"),
            domain_concept=load_prompt("docs", "domain_concept.prompt"),
            base_agent_dev_doc=jinja2.Template(load_prompt("docs", "base_agent_dev.prompt")).render(
                core_scene_agent_module_reference=load_prompt("docs", "references", "core_scene_agent_module.prompt"),
                data_message_module_reference=load_prompt("docs", "references", "data_message_module.prompt"),
            ),
            static_agent_module_name="examiner",
            static_agent_example_code=load_prompt("examples", "base_agents", "examiner.py"),
            dynamic_agent_module_name="examinee",
            dynamic_agent_example_code=load_prompt("examples", "base_agents", "examinee.py"),
            scene_definition_code=target_scene_definition,
        )

        _finished = False

        @function
        def _complete(s, initial_msg):
            nonlocal _finished

            add_and_show_message(s, user, initial_msg, False)
            add_and_show_message(
                s, system, "请你首先根据 scene_definition.py 的信息提取出所有角色名，并判断他们是动态还是静态的角色"
            )
            add_and_show_message(s, assistant, gen("retrieve_role_info", max_tokens=512, temperature=0.0), False)
            add_and_show_message(s, system, "接下来，你需要依次生成各角色的智能体模块代码")
            while True:
                add_and_show_message(
                    s, system, "请你明确现在你要生成的是哪个角色的智能体模块代码？直接给我角色名，不要说其他的话。"
                )
                add_and_show_message(s, assistant, gen("role_name", max_tokens=16, temperature=0.1), False)
                role_name = s.get_var("role_name")
                add_and_show_message(
                    s,
                    system,
                    "首先告知用户你接下来要生成的角色是哪个，然后另起一行，根据用户指定的格式，结合"
                    f" scene_definition.py 中对于角色 {role_name} 的定义，"
                    "依据用户给你的指南并参考示例代码，生成该角色的智能体模块代码，直接给出代码，并在代码生成完毕后向用户征询改进建议",
                )
                add_and_show_message(s, assistant, gen(f"{role_name}.py", max_tokens=4096, temperature=0.5))
                while True:
                    add_and_show_message(s, user, input())
                    add_and_show_message(
                        s, system, "请你根据用户的反馈判断是否需要修改代码实现？直接回答“是”或“否”，不要说其他的话"
                    )
                    add_and_show_message(s, assistant, gen("judgement", max_tokens=2, temperature=0.0), False)
                    if "否" in s.get_var("judgement"):
                        break
                    add_and_show_message(
                        s,
                        system,
                        f"请你结合用户的反馈修改 {role_name}.py 代码，直接给出代码，并在代码生成完毕后再次向用户征询改进建议",
                    )
                    add_and_show_message(s, assistant, gen(f"{role_name}.py", max_tokens=4096, temperature=0.5))
                extract_code_and_save(
                    s.get_var(f"{role_name}.py"), role_name, os.path.join(self.config.target_pkg_path, "agents")
                )
                add_and_show_message(
                    s,
                    system,
                    "请你判断你是否已经生成完 scene_definition.py"
                    " 中定义的全部角色的智能体模块代码？直接回答“是”或“否”，不要说其他的话",
                )
                add_and_show_message(s, assistant, gen("judgement", max_tokens=2, temperature=0.0), False)
                if "是" in s.get_var("judgement"):
                    break

            _finished = True

        _complete.run(initial_msg, stream=True)

        while not _finished:
            time.sleep(0.01)

    def _scene_completion(self):
        target_scene_definition = load_code(os.path.join(self.config.target_pkg_path, "scene_definition.py"), None)
        if not target_scene_definition:
            raise NotImplementedError("must implement scene_definition.py before complete agents' code")

        agent_modules = []
        for root, dirs, files in os.walk(os.path.join(self.config.target_pkg_path, "agents")):
            for file in files:
                if file.endswith(".py") and "__init__" not in file:
                    agent_modules.append(os.path.join(root, file))
        if not agent_modules:
            raise NotImplementedError("must implement all agents base class that defined in scene_definition.py")

        initial_msg = jinja2.Template(load_prompt("system", "scene_developer.prompt")).render(
            leaf_playground_overview=load_prompt("docs", "overview.prompt"),
            domain_concept=load_prompt("docs", "domain_concept.prompt"),
            scene_dev_doc=jinja2.Template(load_prompt("docs", "scene.prompt")).render(
                core_scene_module_reference=load_prompt("docs", "references", "core_scene_module.prompt"),
                data_log_module_reference=load_prompt("docs", "references", "data_log_module.prompt"),
            ),
            scene_example_code=load_code(
                reference_file=(
                    None
                    if not self.config.reference_pkg_path
                    else os.path.join(self.config.reference_pkg_path, "scene.py")
                ),
                default=None if self.config.reference_pkg_path else load_prompt("examples", "scene.py"),
            ),
            scene_definition_code=target_scene_definition,
            agents_code="\n\n".join([
                f"```python\n# {agent_module}\n\n{load_code(agent_module, None)}\n```"
                for agent_module in agent_modules
            ]),
        )

        _finished = False

        @function
        def _complete(s, initial_msg):
            nonlocal _finished

            add_and_show_message(s, user, initial_msg, False)
            add_and_show_message(
                s,
                system,
                "请你首先根据 scene_definition.py 的信息，自行描述一遍完整的场景运行流程，并询问用户是否需要调整",
            )
            add_and_show_message(s, assistant, gen("scene_flow_desc", max_tokens=4096, temperature=0.5))
            while True:
                add_and_show_message(s, user, input())
                add_and_show_message(
                    s,
                    system,
                    "请你根据用户的反馈判断是否需要调整场景运行流程的描述？直接回答“是”或“否”，不要说其他的话",
                )
                add_and_show_message(s, assistant, gen("judgement", max_tokens=2, temperature=0.0), False)
                if "否" in s.get_var("judgement"):
                    break
                add_and_show_message(
                    s, system, "请你结合用户的反馈调整场景运行流程的描述，并再次询问用户是否需要进一步地调整"
                )
                add_and_show_message(s, assistant, gen("scene_flow_desc", max_tokens=4096, temperature=0.5))
            add_and_show_message(
                s,
                system,
                f"请你根据用户指定的格式，结合上面的信息，依据用户给你的指南并参考示例代码，生成 scene.py 代码，"
                f"直接给出代码，并在代码生成完毕后向用户征询改进建议",
            )
            add_and_show_message(s, assistant, gen("scene.py", max_tokens=4096, temperature=0.5))
            while True:
                add_and_show_message(s, user, input())
                add_and_show_message(
                    s, system, "请你根据用户的反馈判断是否需要修改代码实现？直接回答“是”或“否”，不要说其他的话"
                )
                add_and_show_message(s, assistant, gen("judgement", max_tokens=2, temperature=0.0), False)
                if "否" in s.get_var("judgement"):
                    break
                add_and_show_message(
                    s,
                    system,
                    "请你结合用户的反馈修改 scene.py 代码，直接给出代码，并在代码生成完毕后再次向用户征询改进建议",
                )
                add_and_show_message(s, assistant, gen("scene.py", max_tokens=4096, temperature=0.5))
            extract_code_and_save(s.get_var("scene.py"), "scene", self.config.target_pkg_path)

            _finished = True

        _complete.run(initial_msg, stream=True)

        while not _finished:
            time.sleep(0.01)

    def run(self):
        if self.config.definition_completion:
            self._definition_completion()
        if self.config.agent_completion:
            self._agent_completion()
        if self.config.scene_completion:
            self._scene_completion()


__all__ = ["Pipeline", "PipelineConfig"]



from typing import Optional
from uuid import uuid4

from pydantic import Field

from .base import Data


class Role(Data):
    name: str = Field(default=...)
    description: str = Field(default=...)
    is_static: bool = Field(default=False)


class Profile(Data):
    id: str = Field(default_factory=lambda: "agent_" + uuid4().hex[:8])
    name: str = Field(default=...)
    role: Optional[Role] = Field(default=None)


__all__ = ["Role", "Profile"]



import os
from typing import Any, Literal, Optional, Union

import openai
from pydantic import Field

from .base import AIBackend, AIBackendConfig


COMPLETION_MODELS = [
    "babbage-002",
    "davinci-002",
    "gpt-3.5-turbo-instruct",
    "text-davinci-003",
    "text-davinci-002",
    "text-davinci-001",
    "code-davinci-002",
    "text-curie-001",
    "text-babbage-001",
    "text-ada-001",
]
CHAT_MODELS = [
    "gpt-4-1106-preview",
    "gpt-4-vision-preview",
    "gpt-4",
    "gpt-4-0314",
    "gpt-4-0613",
    "gpt-4-32k",
    "gpt-4-32k-0314",
    "gpt-4-32k-0613",
    "gpt-3.5-turbo",
    "gpt-3.5-turbo-16k",
    "gpt-3.5-turbo-0301",
    "gpt-3.5-turbo-0613",
    "gpt-3.5-turbo-16k-0613",
]


_ValidModels = Literal[
    "babbage-002",
    "davinci-002",
    "gpt-3.5-turbo-instruct",
    "text-davinci-003",
    "text-davinci-002",
    "text-davinci-001",
    "code-davinci-002",
    "text-curie-001",
    "text-babbage-001",
    "text-ada-001",
    "gpt-4-1106-preview",
    "gpt-4-vision-preview",
    "gpt-4",
    "gpt-4-0314",
    "gpt-4-0613",
    "gpt-4-32k",
    "gpt-4-32k-0314",
    "gpt-4-32k-0613",
    "gpt-3.5-turbo",
    "gpt-3.5-turbo-16k",
    "gpt-3.5-turbo-0301",
    "gpt-3.5-turbo-0613",
    "gpt-3.5-turbo-16k-0613",
]


class OpenAIBackendConfig(AIBackendConfig):
    model: _ValidModels = Field(default=...)
    api_key: Optional[str] = Field(default=None)
    organization: Optional[str] = Field(default=None)
    base_url: Optional[str] = Field(default=None)
    azure_endpoint: Optional[str] = Field(default=None)
    azure_deployment: Optional[str] = Field(default=None)
    api_version: Optional[str] = Field(default=None)
    is_azure: bool = Field(default=False)
    max_retries: int = Field(default=2)
    timeout: float = Field(default=60)

    def model_post_init(self, __context: Any) -> None:
        super().model_post_init(__context)
        if not self.api_key:
            self.api_key = os.environ.get("AZURE_OPENAI_API_KEY" if self.is_azure else "OPENAI_API_KEY", None)
        if not self.api_key and not self.is_azure:
            raise ValueError(
                "Must provide the `api_key` argument, or the `OPENAI_API_KEY` environment variable when is_azure=False"
            )
        if not self.organization and not self.is_azure:
            self.organization = os.environ.get("OPENAI_ORG_ID", None)
        if not self.base_url and self.is_azure:
            if not self.azure_endpoint:
                self.azure_endpoint = os.environ.get("AZURE_OPENAI_ENDPOINT", None)
            if self.azure_endpoint is None:
                raise ValueError(
                    "Must provide one of the `base_url` or `azure_endpoint` arguments, "
                    "or the `AZURE_OPENAI_ENDPOINT` environment variable when is_azure=True"
                )
        if self.base_url and self.azure_endpoint and self.is_azure:
            raise ValueError("base_url and azure_endpoint are mutually exclusive when is_azure=True")

    @property
    def client_init_payload(self):
        payload = {
            "api_key": self.api_key,
            "organization": self.organization,
            "base_url": self.base_url,
            "timeout": self.timeout,
            "max_retries": self.max_retries,
        }
        if self.is_azure:
            payload.update(
                azure_endpoint=self.azure_endpoint,
                azure_deployment=self.azure_deployment,
                api_version=self.api_version,
            )
        return payload

    def creat_client(self) -> Union[openai.OpenAI, openai.AzureOpenAI]:
        obj = openai.OpenAI if not self.is_azure else openai.AzureOpenAI
        return obj(**self.client_init_payload)

    def creat_async_client(self) -> Union[openai.AsyncOpenAI, openai.AsyncAzureOpenAI]:
        obj = openai.AsyncOpenAI if not self.is_azure else openai.AsyncAzureOpenAI
        return obj(**self.client_init_payload)


class OpenAIBackend(AIBackend):
    config_cls = OpenAIBackendConfig
    config: config_cls

    def __init__(self, config: config_cls):
        super().__init__(config=config)
        self.client = self.config.creat_client()
        self.async_client = self.config.creat_async_client()


__all__ = ["COMPLETION_MODELS", "CHAT_MODELS", "OpenAIBackendConfig", "OpenAIBackend"]



import asyncio
import sys
from functools import partial
from typing import Any, Callable

import anyio

if sys.version_info >= (3, 10):  # pragma: no cover
    from typing import ParamSpec
else:  # pragma: no cover
    from typing_extensions import ParamSpec


P = ParamSpec("P")


async def run_asynchronously(func: Callable[P, Any], *args: P.args, **kwargs: P.kwargs):
    if asyncio.iscoroutinefunction(func):
        return await func(*args, **kwargs)
    else:
        if kwargs:
            func = partial(func, **kwargs)
        return await anyio.to_thread.run_sync(func, *args)


__all__ = ["run_asynchronously"]



from typing import List, Literal, Union
from typing_extensions import Annotated

from leaf_playground.core.scene_definition import *
from leaf_playground.core.scene_definition.definitions.metric import _RecordData
from leaf_playground.data.message import TextMessage
from leaf_playground.data.profile import Profile
from pydantic import Field

from .dataset_util import DatasetConfig


def accuracy_fn(records: List[_RecordData]) -> AggregationMethodOutput:
    num_records = len(records)
    num_accurate = len([record for record in records if bool(record.value)])
    accuracy = round(num_accurate / num_records, 8)
    return AggregationMethodOutput(value=accuracy)


class ExaminerSample(TextMessage):
    sample_id: int = Field(default=...)
    msg_type: Literal["sample"] = Field(default="sample")


class ExamineeAnswer(TextMessage):
    sample_id: int = Field(default=...)
    msg_type: Literal["answer"] = Field(default="answer")


MessageType = Annotated[Union[ExaminerSample, ExamineeAnswer], Field(discriminator="msg_type")]

SCENE_DEFINITION = SceneDefinition(
    name="Mmlu",
    description="using MMLU dataset to test and evaluate agents.",
    roles=[
        RoleDefinition(
            name="examiner",
            description="the one who prepares samples and sends to examinees.",
            num_agents_range=(1, 1),
            is_static=True,
            actions=[
                ActionDefinition(
                    name="prepare_samples",
                    description="prepare samples based on dataset config.",
                    signature=ActionSignatureDefinition(
                        parameters=[ActionSignatureParameterDefinition(name="ds_config", annotation=DatasetConfig)],
                        return_annotation=None,
                    ),
                ),
                ActionDefinition(
                    name="send_sample",
                    description="pick one sample and broadcast to all examinees.",
                    signature=ActionSignatureDefinition(
                        parameters=[ActionSignatureParameterDefinition(name="receivers", annotation=List[Profile])],
                        return_annotation=ExaminerSample,
                    ),
                ),
            ],
        ),
        RoleDefinition(
            name="examinee",
            description="the one who receives samples sent by examiner and answer to those samples.",
            num_agents_range=(1, -1),
            is_static=False,
            actions=[
                ActionDefinition(
                    name="answer",
                    description="answering the question sent by examiner",
                    signature=ActionSignatureDefinition(
                        parameters=[
                            ActionSignatureParameterDefinition(name="sample", annotation=ExaminerSample),
                            ActionSignatureParameterDefinition(name="examiner", annotation=Profile),
                        ],
                        return_annotation=ExamineeAnswer,
                        is_static_method=False,
                    ),
                    metrics=[
                        MetricDefinition(
                            name="accurate",
                            description="accuracy of examinee's answer",
                            record_value_dtype=ValueDType.BOOLEAN,
                            record_display_type=DisplayType.BOOLEAN_RADIO,
                            expect_resp_msg_type=ExamineeAnswer,
                            agg_method=DynamicAggregationFn.create_dynamic_fn(fn=accuracy_fn),
                            is_comparison=False,
                        )
                    ],
                )
            ],
        ),
    ],
    env_vars=[],
)

__all__ = ["ExaminerSample", "ExamineeAnswer", "MessageType", "SCENE_DEFINITION"]



import asyncio
import traceback
from datetime import datetime
from typing import Any, Callable, List, Literal, Optional

from fastapi import HTTPException
from sqlmodel import SQLModel
from sqlalchemy.ext.asyncio import create_async_engine, AsyncSession
from sqlalchemy import select, Select

from leaf_playground._type import Singleton
from leaf_playground.core.scene_engine import SceneEngineState
from leaf_playground.utils.thread_util import run_asynchronously

from .model import *


class DB(Singleton):
    def __init__(self, db_url: str):
        self.url = db_url
        self.engine = create_async_engine(db_url)
        asyncio.ensure_future(self._on_startup())
        asyncio.ensure_future(self._background_job())

        self._run = asyncio.Event()

    async def wait_db_startup(self):
        await self._run.wait()

    async def _on_startup(self):
        async with self.engine.begin() as conn:
            # 创建所有表结构
            await conn.run_sync(SQLModel.metadata.create_all)
        self._run.set()

    async def _background_job(self):
        await self._run.wait()

        deleting_tasks = await self.get_task_by_life_cycle(TaskDBLifeCycle.DELETING)
        for task in deleting_tasks:
            await self.delete_task(task.id)

    async def _insert(self, data: SQLModel) -> int:
        async with AsyncSession(self.engine) as session:
            try:
                session.add(data)
                await session.commit()
                succeed = True
            except:
                succeed = False
        return 200 if succeed else 500

    async def _update(self, obj_getter: Callable[[AsyncSession], Any], **kwargs) -> int:
        async with AsyncSession(self.engine) as session:
            obj = await run_asynchronously(obj_getter, session)
            if not obj:
                return 404
            for k, v in kwargs.items():
                try:
                    setattr(obj, k, v)
                except:
                    return 422
            try:
                session.add(obj)
                await session.commit()
            except:
                return 500
        return 200

    async def insert_task(self, task: Task):
        status_code = await self._insert(TaskTable.from_task(task))
        if status_code != 200:
            raise HTTPException(status_code=status_code, detail=f"insert task [{task.id}] failed.")

    async def update_task_status(self, tid: str, status: str):
        async def obj_getter(session: AsyncSession):
            return await session.get(TaskTable, tid)

        valid_status = [s.value for s in SceneEngineState]
        if status not in valid_status:
            raise HTTPException(status_code=422, detail=f"{status} is not valid, valid status are {valid_status}")

        status_code = await self._update(obj_getter, **{"status": status})
        if status_code == 404:
            raise HTTPException(status_code=status_code, detail=f"task [{tid}] not found")
        if status_code == 422:
            raise HTTPException(status_code=status_code, detail=f"task [{tid}] update failed")
        if status_code == 500:
            raise HTTPException(
                status_code=500,
                detail=f"task [{tid}] update succeeded but session commit failed, maybe try again later",
            )

    async def get_task(self, tid: str) -> Task:
        async with AsyncSession(self.engine) as session:
            task = await session.get(TaskTable, tid)
            if not task:
                raise HTTPException(status_code=404, detail=f"task [{tid}] not found")
        return task.to_task()

    async def get_task_by_life_cycle(self, life_cycle: TaskDBLifeCycle) -> List[Task]:
        async with AsyncSession(self.engine) as session:
            tasks = (
                (await session.execute(select(TaskTable).where(TaskTable.live_cycle == life_cycle))).scalars().all()
            )
        if not tasks:
            return []
        return [task.to_task() for task in tasks]

    async def delete_task(self, tid: str):
        async def _delete_logs_batchify(session: AsyncSession):
            while True:
                try:
                    result = await session.execute(select(LogTable).where(LogTable.tid == tid).limit(100))
                    logs = result.scalars().all()
                    if not logs:
                        break
                    for log in logs:
                        await session.delete(log)
                    await session.commit()
                except:
                    await session.rollback()
                    raise

        async def _delete_messages_batchify(session: AsyncSession):
            while True:
                try:
                    result = await session.execute(select(MessageTable).where(MessageTable.tid == tid).limit(100))
                    messages = result.scalars().all()
                    if not messages:
                        break
                    for message in messages:
                        await session.delete(message)
                    await session.commit()
                except:
                    await session.rollback()
                    raise

        async def _delete_task():
            session = AsyncSession(self.engine)
            task = await session.get(TaskTable, tid)
            if not task:
                await session.close()
                return

            # update task live cycle to "deleting"
            try:
                if task.live_cycle != TaskDBLifeCycle.DELETING:
                    task.live_cycle = TaskDBLifeCycle.DELETING
                    await session.commit()
            except:
                traceback.print_exc()
                await session.rollback()
                await session.close()
                raise
            try:
                # delete logs first
                await _delete_logs_batchify(session)
                # then delete messages
                await _delete_messages_batchify(session)
            except:
                traceback.print_exc()
                await session.close()
                raise
            # finally delete the task itself
            try:
                await session.delete(task)
                await session.commit()
            except:
                traceback.print_exc()
                await session.rollback()
                await session.close()
                raise

        await _delete_task()  # TODO: using aio-pika to do delete in truly background

    async def insert_log(self, log: Log):
        status_code = await self._insert(LogTable.from_log(log))
        if status_code != 200:
            raise HTTPException(status_code=status_code, detail=f"insert log [{log.id}] failed")

    async def update_log(self, new_log: Log):
        log_id = new_log.id

        async def obj_getter(session: AsyncSession):
            return await session.get(LogTable, log_id)

        status_code = await self._update(obj_getter, **new_log.model_dump())
        if status_code == 404:
            raise HTTPException(status_code=status_code, detail=f"log [{log_id}] not found")
        if status_code == 422:
            raise HTTPException(status_code=status_code, detail=f"log [{log_id}] update failed")
        if status_code == 500:
            raise HTTPException(
                status_code=500,
                detail=f"log [{log_id}] update succeeded but session commit failed, maybe try again later",
            )

    async def get_log_by_id(self, log_id: str) -> Log:
        async with AsyncSession(self.engine) as session:
            log = await session.get(LogTable, log_id)
            if not log:
                raise HTTPException(status_code=404, detail=f"log [{log_id}] not found")
        return log.to_log()

    async def get_logs_by_tid_with_update_time_constraint(
        self, tid: str, last_checked_dt: Optional[datetime] = None
    ) -> List[Log]:
        if not last_checked_dt:
            statement: Select = select(LogTable).where(LogTable.tid == tid)
        else:
            statement: Select = select(LogTable).where(LogTable.tid == tid, LogTable.db_last_update > last_checked_dt)
        statement = statement.order_by(LogTable.db_last_update if last_checked_dt is not None else LogTable.created_at)
        async with AsyncSession(self.engine) as session:
            logs = (await session.execute(statement)).scalars().all()
        if not logs:
            return []
        return [log.to_log() for log in logs]

    async def insert_message(self, message: Message):
        status_code = await self._insert(MessageTable.from_message(message))
        if status_code != 200:
            raise HTTPException(status_code=status_code, detail=f"insert message [{message.id}] failed")

    async def get_message_by_id(self, message_id: str) -> Message:
        async with AsyncSession(self.engine) as session:
            message = await session.get(MessageTable, message_id)
            if not message:
                raise HTTPException(status_code=404, detail=f"message [{message_id}] not found")
        return message.to_message()


__all__ = ["DB"]



import re
from typing import List

from pydantic import Field

from .base import EvalTool, EvalToolConfig


class RegexEvalToolConfig(EvalToolConfig):
    regex_rules: List[str] = Field(
        default=[r"\b([A-D])\b"], description="Regex rules for extracting answers from the scene."
    )
    ignore_case: bool = Field(
        default=True, description="Whether to ignore case when matching regex rules and answers."
    )


class RegexEvalTool(EvalTool):
    config_obj = RegexEvalToolConfig
    config: config_obj

    def __init__(self, config: config_obj):
        super().__init__(config)
        self._regex_rules = config.regex_rules
        self._ignore_case = config.ignore_case

    def extract_answer(self, origin_answer: str) -> str:
        regex_flag = re.IGNORECASE if self.ignore_case else 0
        for rule in self._regex_rules:
            match = re.search(rule, origin_answer, flags=regex_flag)
            if match:
                return match.group(1)
        return ""

    @property
    def ignore_case(self) -> bool:
        return self._ignore_case



from sys import _getframe
from typing import Any, Dict, List, Literal, Optional, Tuple, Type, Union

from pydantic import create_model, BaseModel, Field, PositiveInt, PrivateAttr

from . import MetricDefinition
from .definitions import EnvVarDefinition, EnvVarConfig, MetricConfig, RoleDefinition, RoleConfig
from ..._config import _Config

_EnvVarName = str
_RoleName = str


class SceneDefinition(BaseModel):
    name: str = Field(default=...)
    description: str = Field(default=...)
    env_vars: List[EnvVarDefinition] = Field(default=...)
    roles: List[RoleDefinition] = Field(default=...)

    _log_exporters: List["leaf_playground.core.workers.logger.Logger"] = PrivateAttr(default=[])

    @property
    def roles_agent_num_range(self) -> Dict[str, Tuple[PositiveInt, Union[PositiveInt, Literal[-1]]]]:
        return {role.name: role.num_agents_range for role in self.roles}

    @property
    def log_exporters(self):
        return self._log_exporters

    def model_post_init(self, __context: Any) -> None:
        from leaf_playground.core.workers.logger import _KeptLogExporter

        if len(set([r.name for r in self.roles])) != len(self.roles):
            raise ValueError(f"roles should have unique names")
        if len(set([e.name for e in self.env_vars])) != len(self.env_vars):
            raise ValueError(f"env_vars should have unique names")

        for ext in ["json", "jsonl", "csv"]:
            self._log_exporters.append(_KeptLogExporter(extension=ext))

    def get_role_definition(self, role_name: str) -> RoleDefinition:
        for role in self.roles:
            if role.name == role_name:
                return role
        raise ValueError(f"role [{role_name}] not found")

    def get_env_var_definition(self, env_var_name: str) -> EnvVarDefinition:
        for env_var in self.env_vars:
            if env_var.name == env_var_name:
                return env_var
        raise ValueError(f"env_var [{env_var_name}] not found")

    def get_metric_definition(self, metric_belonged_chain: str) -> MetricDefinition:
        role_name, action_name, metric_name = metric_belonged_chain.split(".")
        return (
            self.get_role_definition(role_name).get_action_definition(action_name).get_metric_definition(metric_name)
        )

    def registry_log_exporters(self, log_exporters: List["leaf_playground.core.workers.logger.LogExporter"]):
        name2exporter = {f"{exporter.file_name}.{exporter.extension}": exporter for exporter in self.log_exporters}
        for exporter in log_exporters:
            name = f"{exporter.file_name}:{exporter.extension}"
            name2exporter[name] = exporter

        self._log_exporters = list(name2exporter.values())


class SceneEnvVarsConfig(_Config):
    @classmethod
    def create_config_model(cls, scene_definition: SceneDefinition) -> Optional[Type["SceneEnvVarsConfig"]]:
        model_name = "".join([each.capitalize() for each in scene_definition.name.split("_")]) + cls.__name__
        module = _getframe(1).f_globals["__name__"]
        fields = {}
        for env_var in scene_definition.env_vars:
            fields[env_var.name] = (EnvVarConfig.create_config_model(env_var), Field(default=...))
        if not fields:
            return
        return create_model(__model_name=model_name, __module__=module, __base__=cls, **fields)

    def get_env_var_config(self, env_var_name: str) -> EnvVarConfig:
        return getattr(self, env_var_name)

    def initiate_env_vars(self) -> Dict[_EnvVarName, "leaf_playground.data.environment.EnvironmentVariable"]:
        return {f_name: getattr(self, f_name).initiate_env_var() for f_name in self.model_fields_set}


class SceneRolesConfig(_Config):
    @classmethod
    def create_config_model(cls, scene_definition: SceneDefinition) -> Type["SceneRolesConfig"]:
        model_name = "".join([each.capitalize() for each in scene_definition.name.split("_")]) + cls.__name__
        module = _getframe(1).f_globals["__name__"]
        fields = {}
        for role in scene_definition.roles:
            fields[role.name] = (RoleConfig.create_config_model(role), Field(default=...))
        return create_model(__model_name=model_name, __module__=module, __base__=cls, **fields)

    def get_role_config(self, role_name: str) -> RoleConfig:
        return getattr(self, role_name)

    def initiate_agents(self) -> Dict[_RoleName, List["leaf_playground.core.scene_agent.SceneAgent"]]:
        return {f_name: getattr(self, f_name).initiate_agents() for f_name in self.model_fields_set}


class SceneConfig(_Config):
    env_vars_config: Optional[SceneEnvVarsConfig] = Field(default=...)
    roles_config: SceneRolesConfig = Field(default=...)

    @classmethod
    def create_config_model(
        cls,
        scene_definition: SceneDefinition,
        additional_config_fields: Optional[Dict[str, Tuple[Type, Field]]] = None,
    ) -> Type["SceneConfig"]:
        model_name = "".join([each.capitalize() for each in scene_definition.name.split("_")]) + cls.__name__
        module = _getframe(1).f_globals["__name__"]
        fields = {}
        env_vars_config_model = SceneEnvVarsConfig.create_config_model(scene_definition)
        if env_vars_config_model:
            fields["env_vars_config"] = (env_vars_config_model, Field(default=...))
        else:
            fields["env_vars_config"] = (Literal[None], Field(default=None))
        fields["roles_config"] = (SceneRolesConfig.create_config_model(scene_definition), Field(default=...))
        if additional_config_fields:
            fields.update(**additional_config_fields)

        return create_model(__model_name=model_name, __module__=module, __base__=cls, **fields)

    def get_env_var_config(self, env_var_name: str) -> EnvVarConfig:
        return self.env_vars_config.get_env_var_config(env_var_name)

    def get_role_config(self, role_name: str) -> RoleConfig:
        return self.roles_config.get_role_config(role_name)

    def get_metric_config(self, metric_belonged_chain: str) -> MetricConfig:
        role_name, action_name, metric_name = metric_belonged_chain.split(".")
        return self.get_role_config(role_name).get_action_config(action_name).get_metric_config(metric_name)

    def init_env_vars(self) -> Dict[_EnvVarName, "leaf_playground.data.environment.EnvironmentVariable"]:
        return self.env_vars_config.initiate_env_vars() if self.env_vars_config else {}

    def init_agents(self) -> Dict[_RoleName, List["leaf_playground.core.scene_agent.SceneAgent"]]:
        return self.roles_config.initiate_agents()


__all__ = ["SceneDefinition", "SceneEnvVarsConfig", "SceneRolesConfig", "SceneConfig"]



import os.path
from contextlib import asynccontextmanager
from typing import Literal

from fastapi import FastAPI, Depends
from fastapi.responses import JSONResponse
from pydantic import BaseModel, Field, DirectoryPath

from leaf_playground import __version__ as leaf_version

from .hub import *
from .task import *
from .task.model import TaskRunTimeEnv


class AppConfig(BaseModel):
    hub_dir: DirectoryPath = Field(default=...)
    server_port: int = Field(default=...)
    server_host: str = Field(default=...)
    runtime_env: TaskRunTimeEnv = Field(default=...)


class AppInfo(BaseModel):
    name: Literal["leaf-playground-local-server"] = Field(default="leaf-playground-local-server")
    version: str = Field(default=leaf_version)
    hub_dir: str = Field(default=...)


app_config: AppConfig = None
app_info: AppInfo = None


def config_server(config: AppConfig):
    global app_config, app_info
    app_config = config
    app_info = AppInfo(hub_dir=app_config.hub_dir.as_posix())

    os.makedirs(os.path.join(app_config.hub_dir, ".leaf_workspace"), exist_ok=True)


@asynccontextmanager
async def lifespan(app: FastAPI):
    db = DB(f"sqlite+aiosqlite:///{os.path.join(app_config.hub_dir, '.leaf_workspace', '.leaf_playground.db')}")
    await db.wait_db_startup()
    Hub(hub_dir=app_config.hub_dir.as_posix())
    TaskManager(server_port=app_config.server_port, runtime_env=app_config.runtime_env)

    try:
        yield
    except:
        pass


app = FastAPI(title="leaf-playground-local-server", version=leaf_version, lifespan=lifespan)
app.include_router(hub_router)
app.include_router(task_router)


@app.get("/", response_class=JSONResponse)
async def main_page(hub: Hub = Depends(Hub.get_instance)) -> JSONResponse:
    """response is a json dict with two fields: projects(simple information for projects),
    app_info(metadata of the server)"""
    projects = []
    for proj in hub.projects.values():
        projects.append({
            "name": proj.name,
            "display_name": proj.metadata.scene_metadata["scene_definition"]["name"],
            "id": proj.id,
            "description": proj.metadata.scene_metadata["scene_definition"]["description"],
        })
    return JSONResponse(content={"projects": projects, "app_info": app_info.model_dump(mode="json")})


__all__ = ["config_server", "AppConfig"]



from inspect import Signature, Parameter
from warnings import warn
from sys import _getframe
from typing import Any, Annotated, List, Literal, Optional, Type, Union

from pydantic import create_model, field_serializer, BaseModel, Field, PrivateAttr

from .metric import MetricConfig, MetricDefinition
from ...._config import _Config
from ....data.message import Message


class ActionSignatureParameterDefinition(BaseModel):
    name: str = Field(default=...)
    annotation: Optional[Any] = Field(default=Parameter.empty)

    @field_serializer("annotation")
    def serialize_annotation(self, annotation: Optional[Type], _info):
        return str(annotation)


class ActionSignatureDefinition(BaseModel):
    parameters: Optional[List[ActionSignatureParameterDefinition]] = Field(default=None)
    return_annotation: Optional[Type[Message]] = Field(default=Signature.empty)
    is_static_method: bool = Field(default=False)

    def model_post_init(self, __context: Any) -> None:
        if self.return_annotation not in [None, Signature.empty] and not issubclass(self.return_annotation, Message):
            raise TypeError("return_annotation must be a subclass of Message if is explicitly specified")

    @field_serializer("return_annotation")
    def serialize_return_annotation(self, return_annotation: Optional[Type[Message]], _info):
        return str(return_annotation)


class ActionDefinition(BaseModel):
    name: str = Field(default=...)
    description: str = Field(default=...)
    signature: ActionSignatureDefinition = Field(default=...)
    metrics: Optional[List[MetricDefinition]] = Field(default=None)
    exec_timeout: int = Field(default=30)

    _belonged_role: Optional["leaf_playground.core.scene_info.definitions.role.RoleDefinition"] = PrivateAttr(
        default=None
    )

    @property
    def belonged_role(self):
        return self._belonged_role

    @property
    def belonged_chain(self):
        return self.belonged_role.name + "." + self.name

    def get_metric_definition(self, metric_name: str) -> MetricDefinition:
        for metric in self.metrics or []:
            if metric.name == metric_name:
                return metric
        raise ValueError(f"metric [{metric_name}] not found")

    def model_post_init(self, __context: Any) -> None:
        if self.signature.return_annotation in [None, Signature.empty] and self.metrics:
            warn(
                f"action [{self.name}] hasn't output, but metrics are specified, "
                "this may cause some unexpected behaviors"
            )
        if self.metrics and len(set([m.name for m in self.metrics])) != len(self.metrics):
            raise ValueError(f"metrics of action [{self.name}] should have unique names")
        for metric in self.metrics or []:
            if metric.belonged_action:
                raise ValueError(
                    f"[{metric.name}] metric has already been bounded to action [{metric.belonged_action.name}]"
                )
            metric._belonged_action = self
            if metric.expect_resp_msg_type != self.signature.return_annotation:
                raise TypeError(
                    f"metric [{metric.name}] expects using {metric.expect_resp_msg_type.__name__} to evaluate, "
                    f"but action [{self.name}]'s result type is {self.signature.return_annotation.__name__}"
                )

        if not self.signature.is_static_method:
            self.signature.parameters = (
                [ActionSignatureParameterDefinition(name="self")]
                if not self.signature.parameters
                else [ActionSignatureParameterDefinition(name="self")] + self.signature.parameters
            )

    def get_signature(self) -> Signature:
        return Signature(
            parameters=[
                Parameter(name=p.name, kind=Parameter.POSITIONAL_OR_KEYWORD, annotation=p.annotation)
                for p in (self.signature.parameters or [])
            ],
            return_annotation=self.signature.return_annotation,
        )


class ActionMetricsConfig(_Config):
    @classmethod
    def create_config_model(cls, action_definition: ActionDefinition) -> Optional[Type["ActionMetricsConfig"]]:
        if not action_definition.metrics:
            return

        model_name = "".join([each.capitalize() for each in action_definition.name.split("_")]) + cls.__name__
        module = _getframe(1).f_globals["__name__"]
        fields = {}
        for metric in action_definition.metrics:
            fields[metric.name] = (MetricConfig, Field(default=...))
        return create_model(__model_name=model_name, __module__=module, __base__=cls, **fields)

    def get_metric_config(self, metric_name: str) -> MetricConfig:
        return getattr(self, metric_name)


class ActionConfig(_Config):
    metrics_config: Optional[ActionMetricsConfig] = Field(default=...)

    @classmethod
    def create_config_model(cls, action_definition: ActionDefinition) -> Type["ActionConfig"]:
        model_name = "".join([each.capitalize() for each in action_definition.name.split("_")]) + cls.__name__
        module = _getframe(1).f_globals["__name__"]
        metrics_config_model = ActionMetricsConfig.create_config_model(action_definition)

        fields = {}
        if not metrics_config_model:
            fields["metrics_config"] = (Literal[None], Field(default=None))
        else:
            fields["metrics_config"] = (metrics_config_model, Field(default=...))
        return create_model(__model_name=model_name, __module__=module, __base__=cls, **fields)

    def get_metric_config(self, metric_name: str):
        return self.metrics_config.get_metric_config(metric_name)


__all__ = [
    "ActionSignatureParameterDefinition",
    "ActionSignatureDefinition",
    "ActionDefinition",
    "ActionMetricsConfig",
    "ActionConfig",
]
